/* Uses skorch docs theme from https://skorch.readthedocs.io */

.wy-side-nav-search {
	background-color: #3B556F;
	#background-color: #fcfcfc;
}

.wy-side-nav-search > div.version {
	color: #F8F9F9;
	#color: black;
}

.wy-side-nav-search > a {
	color: #F8F9F9;
	#color: black;
}

/* Adds logo in header for mobile */

.wy-nav-top {
	background-color: #3B556F;
	#background-color: #fcfcfc;
	background-image: url('../logo.png');
	background-repeat: no-repeat;
	background-position: bottom;
	padding: 0;
	margin: 0.4045em 0.809em;
	color: #333;
}

.wy-nav-top > a {
	display: none;
}
